/*
got game player script.
created entirely by dave caruso; may 2021

THIS FILE HANDLES A LOT AND IT IS A MESS. A LOT OF THIS IS HACKED TOGETHER AND
IS BAD CODE, BUT IT WORKS.

ALL COMMENTS WERE WRITTEN AFTER, WHERE I TRY AND UNDERSTAND WHAT I WROTE

*/
import * as c from './constants.js'
import * as CameraLogic from './camera.js';
import { getTargetAngle, step, applyVel, applyAccel, lerp, intersects, offsetRect, isPressingJump, isPressingUnpause, isPressingRotateLeft, isPressingRotateRight, key } from './util.js'

// global store for the player object. its a Sprite with these extra props
// 		- `player.state` is "paused" | "moving" | "dead"
// 		- `player.vy` is y velocity
// 		- `player.angleMode` is what rotation speed it is using (for the mini jumps when you are paused)
//                           aka "normal" | "left" | "right" | "dead"
//                               (180deg)   (-90deg) (90deg)   (special)
// 		- `player.failedColor` is color for visual to show when the color mismatches.

let player;

// death time delay. goes from 0 to .6 when you are in the dead state
let timeDead = 0;

const rotateTimings = {
	normal: c.ROTATE_TIME,
	left: -c.ROTATE_TIME_SM,
	right: c.ROTATE_TIME_SM,
}

// this is a stupid hacky way of getting the localstorage, and also providing a fallback
// the global store of the checkpoint
// { x, y, angle, jump }
const storageId = 'davecode_cube_game_current_checkpoint';
let checkpoint = (() => {
	try {
		return JSON.parse(localStorage.getItem(storageId));
	} catch(e) { return null }
})();

// first and last hardcoded to not
if (checkpoint && (checkpoint.x > 15100 || checkpoint.x < 1100)) {
	checkpoint = null;
}

export function playerInit(newPlayer) {
	player = newPlayer;
	
	// restore a checkpoint if one exists
	if (checkpoint) {
		player.state = 'paused';
		player.vy = 0;
		player.x = checkpoint.x;
		player.y = checkpoint.y;
		player.angle = checkpoint.angle;
	} else {
		player.vy = 0;
		player.y -= 0.001;
		player.state = 'moving';
		player.isVisible = false;
	}
	
	player.width = 64;
	player.height = 64;

	player.angleMode = 'normal';
	player.failedColor = null;
	
	player.playedDeathSound = false;
}

// get collision returns the first object that the player currently is colliding with.
// we will move the player (x,y) temporarily, then move it back, in order to check collisions.
function getCollision() {
	const playerRect = offsetRect(player, -32, -32);
	return [
		...runtime.objects.SolidB.getAllInstances(),
		...runtime.objects.SolidG.getAllInstances(),
		...runtime.objects.Solid.getAllInstances(),
	].find((solid) => intersects(solid, playerRect))
}

// get collision with a pause block. i had to hack this to ignore y, but i didnt feel like rewriting the intersects, so i
// set the player's offset rect to a Stupid large y size
function getCollisionPause() {
	const playerRect = offsetRect(player, -64-32, -32);
	playerRect.y = -2000;
	playerRect.height = 5000;
	return runtime.objects.Pause.getAllInstances().filter(x => !x.used).find((solid) => intersects(solid, playerRect))
}
// same for music markers
function getCollisionMusic() {
	const playerRect = offsetRect(player, -64-1, -32);
	playerRect.y = -2000;
	playerRect.height = 5000;
	return runtime.objects.MusicZone.getAllInstances().filter(x => !x.used).find((solid) => intersects(solid, playerRect))
}

// get collision with kill object, which has an altered hit box
function getCollisionKill() {
	const playerRect = offsetRect(player, -32, -32 - 64 + 20);
	return runtime.objects.Kill.getAllInstances().find((solid) => intersects(solid, playerRect))
}

// get collision with markers. i put these in to hardcode some behavior
function getCollisionMarker(marker) {
	const playerRect = offsetRect(player, -32, -32);
	return runtime.objects[marker].getAllInstances().find((solid) => intersects(solid, playerRect))
}

// get what color name is on the bottom of your cube
function getBottomColor() {
	let target = getTargetAngle(player.angle)
	return (target > (TAU-TAU_8) || target < TAU_2-TAU_8) ? "green" : "blue"
}

// 1 thru 9 keys
function warpCheckpointNumber(n) {
	
	const p = runtime.objects.Pause.getAllInstances().sort((a,b)=>(a.x-b.x))[n - 1];
	if (p.used && player.state === 'paused') return
	
	runtime.objects.Pause.getAllInstances().forEach(x => x.used = false);
	
	p.used = true;
	
	player.state = 'paused';
	player.x = p.x + 32;
	player.y = p.y + 32;
	player.angle = 0;

	runtime.globalVars.ScriptText = p.instVars.Message;

	checkpoint = { x: player.x, y: player.y, angle: player.angle, jump: false  };

	// sync to storage
	localStorage.setItem(storageId, JSON.stringify(checkpoint));
}

// run every tick
export function playerTick() {
	let collision;

	const isJumpAllowed = !getCollisionMarker('NoJumpMarker');
	let isClipping = false;
	
	// [handle death timers]
	if (player.state === 'dead') {
		timeDead += dt;

		if (timeDead > 0.6) {
			player.state = 'paused';
			player.x = checkpoint.x;
			player.y = checkpoint.y;
			player.angle = checkpoint.angle;
			player.failedColor = null;
		}
	} else {
		timeDead = 0;
		player.playedDeathSound = false;
	}

	// [handle x movement]
	if (player.state === 'moving') {
		player.x = applyVel(player.x, c.X_SPEED_PIXELS);
	}
	
	if (getCollisionKill()) {
		player.state = 'dead';
		player.vy = 0;
	}

	// [handle x collision. these cause deaths]
	if ((collision = getCollision()) && !getCollisionMarker('DisableAllCollisionMarker')) {
		// this is the "clipping check"
		// if your velocity is moving enough in one way, the collision is ignored
	    const originalY = player.y;
		player.y += player.y > collision.y ? 15 : player.vy * 0.05;
		if (getCollision()) {
			player.state = 'dead';
			player.vy = 0;
		}
		// restore previous y, and mark clipping on
		player.y = originalY;
		isClipping = true;
	}

    // [handle y acceleration]
	const oldY = player.y;
	
	if (player.state !== 'dead' || player.failedColor) {
		player.y = applyAccel(player.y, player.vy, c.GRAVITY_PIXELS);
		player.vy = applyVel(player.vy, c.GRAVITY_PIXELS);
	}
	
	const yDir = Math.sign(player.vy) || -1;

	let onGround = false;
	
	// [handle y collision]
	while ((!isClipping && (collision = getCollision())) && !getCollisionMarker('DisableAllCollisionMarker')) {
		if (yDir === -1) {
			// ceiling clip part 2
			player.y += 15;
			if (getCollision()) {
				player.state = 'dead';
				player.failedColor = null;
			}
			player.y -= 15;
			isClipping = true;
		} else {
			const type = collision.objectType.name;
			const color = { SolidB: 'blue', SolidG: 'green' }[type] || null;

			// cool trick which snaps us to the top or bottom of the collision object
			// this also supports ceilings but those kill you immediatly. was made for
			// potential anti-gravity like the geometry dash game has.
			player.y = yDir == 1
				? collision.y - 32.001
				: collision.y + collision.height + 32.001;
			// reset velocity
			player.vy = 0;

			// mark ground on
			onGround = true;

			if ((player.state !== 'dead') && color && getBottomColor() !== color) {
				player.state = 'dead';
				player.failedColor = color;
				player.vy = -c.JUMP_VELOCITY_DEATH_PIXELS;
				player.angleMode = "dead";
				//player.angle = step(player.angle, getTargetAngle(player.angle), TAU/5);
				player.angle = getTargetAngle(player.angle);
			}

			if ((player.state !== 'dead') && yDir === -1) {
				player.state = 'dead';
				player.failedColor = null;
			}
		}
	}

    // [handle angle movement]
	// this isnt based on if you're on the ground, but if you *moved*. i do not know why.
	if (player.state !== 'dead' || player.failedColor !== null) {
		if (player.y !== oldY) {
			// rotate based on the set rotate timing. this lets me program the left and right arrows as ACTUAL JUMPS
			// but with different physics. i forget how i calculated the rotate timings but they work out to go either
			// {180 degrees or 90 degrees} per time it takes to jump.
	
			// near the end of the game, i made this more complex with the "dead" angle mode
			const a = player.angleMode === 'dead'
				? timeDead > 0.06 ? 1.4 : -1.2
				: rotateTimings[player.angleMode];
			// ALSO, wrap around the value now, as getTargetAngle does not handle the <0 and >TAU cases.
			player.angle = (player.angle + (TAU_2) * (dt/(a))) % TAU;
		} else {
			// use lerp to smoothly transition to the target angle.
			player.angle = lerp(player.angle, getTargetAngle(player.angle), dt * 40);
			player.angleMode = 'normal';

			// mark on ground ... again. i guess the y collision doesnt always activate???
			onGround = true;
		}
	}

	// [jump handling]
	if (onGround					// normal jumps need the ground
		&& isPressingJump()			// pressing up arrow
		&& checkpoint				// jumping is disallowed during the null checkpoint (start of game)
		&& player.state === 'moving'// you must be moving. when pressing up to unpause, this triggers the frame after.
		&& isJumpAllowed
	) {
		// duh
		player.vy = -c.JUMP_VELOCITY_PIXELS;
		if(getCollisionMarker('JumpBoostMarker')) { player.vy *= 1.05; }
		// the angle is slightly snapped, kinda like a gentle quantization. this helps keep the rotations consistant
		// (A LOT OF TESTING WAS DONE HERE)
		player.angle = step(player.angle, getTargetAngle(player.angle), TAU/2);
	}
	
	// [rotate handling in pause mode]
	let rotateRight = isPressingRotateRight();
	let rotateLeft = isPressingRotateLeft();
	if (onGround						// duh
		&& (rotateRight ^ rotateLeft) 	// XOR; true if ONLY ONE input is true
		&& (player.state === 'paused')	// this can only be done in paused mode
	) {
		// use the small jump
		player.vy = -c.JUMP_VELOCITY_SM_PIXELS;
		// fully snap the angle, i dont want any funny business with this rotate
		player.angle = getTargetAngle(player.angle);
		// mark angle mode accordingly
		player.angleMode = rotateRight ? 'right' : 'left';
	}
	
	// [handle entering pause mode]
	const p = getCollisionPause();
	if (player.state === 'moving' && p) {
		player.state = 'paused';
		runtime.objects.Pause.getAllInstances().forEach(x => x.used = false);
		p.used = true;
		player.x = p.x + 32;
		
		checkpoint = { x: player.x, y: player.y, angle: player.angle, jump: false  };
		
		// sync to storage
		localStorage.setItem(storageId, JSON.stringify(checkpoint));
	}
	
	if (p) {
		runtime.globalVars.ScriptText = p.instVars.Message;
	}
	
	// [handle 1 thru 8 buttons]
	for (let i = 1; i<=8; i++) {
		if (key("Digit" + i)) warpCheckpointNumber(i);
	}
	
	// [handle exiting pause mode]
	if (player.state === 'paused' && isPressingUnpause() && onGround) {
		player.state = 'moving';
		
		// set the checkpoint to match the current angle, and 0also on the ground...
		checkpoint = { x: player.x, y: player.y, angle: getTargetAngle(player.angle), jump: false  };
		
		// sync to storage
		localStorage.setItem(storageId, JSON.stringify(checkpoint));
		
		// [unused feature where checkpoints can auto-jump]
		if (onGround && checkpoint.jump) {
			player.vy = -c.JUMP_VELOCITY_PIXELS;
		 	player.angle = step(player.angle, getTargetAngle(player.angle), TAU/2);
		}
		
		runtime.globalVars.ScriptText = "";
	}
	
	// [handle music]
	if (collision=getCollisionMusic()) {
		runtime.globalVars.ScriptTargetAudio = collision.instVars.ID;
	}
	
	// [handle loop]
	if (collision=getCollisionMarker("Loopback")) {
		player.x = -250;
		CameraLogic.AddX();
	}
	
	// [death audio]
	if (player.state === 'dead' && !player.playedDeathSound) {
		runtime.globalVars.ScriptTriggerSound = player.failedColor ? "wrong" : "crash";
		player.playedDeathSound = true;

		CameraLogic.Shake();
		
		if (!player.failedColor) {
			// ugly code why is it here, whatever
			let a = Math.random()>0.5;
			for (let angle = 0; angle < TAU; angle+=TAU/8) {
				const dp = runtime.objects.DeathParticle.createInstance("Player", player.x, player.y);
				dp.life = 0;
				dp.vx = Math.cos(player.angle + angle);
				dp.vy = Math.sin(player.angle + angle);
				dp.vel = 1;
				dp.scale = 0.5;
				dp.scalev = 0;
				a = !a;
				dp.setAnimation(a ? 'a' : 'b');
			}
		}
	}
	
	// visuals. if i said i had playervisual as a separate object i used to but now i dont
	player.isVisible = player.state !== 'dead' || player.failedColor;
	player.setAnimation(player.failedColor || "normal");
}
